import enum
from typing import Optional, List, Dict, Any


from typing_extensions import Literal
from pydantic import BaseModel, Field

class Sep9Customer(BaseModel):
    last_name: Optional[str]
    first_name: Optional[str]
    additional_name: Optional[str]
    address_country_code: Optional[str]
    state_or_province: Optional[str]
    city: Optional[str]
    postal_code: Optional[str]
    address: Optional[str]
    mobile_number: Optional[str]
    email_address: Optional[str]
    birth_date: Optional[str]
    birth_place: Optional[str]
    birth_country_code: Optional[str]
    bank_account_number: Optional[str]
    bank_number: Optional[str]
    bank_phone_number: Optional[str]
    bank_branch_number: Optional[str]
    tax_id: Optional[str]
    tax_id_name: Optional[str]
    occupation: Optional[str]
    employer_name: Optional[str]
    employer_address: Optional[str]
    language_code: Optional[str]
    id_type: Optional[str]
    id_country_code: Optional[str]
    id_issue_date: Optional[str]
    id_expiration_date: Optional[str]
    id_number: Optional[str]
    photo_id_front: Any
    photo_id_back: Any
    notary_approval_of_photo_id: Any
    ip_address: Optional[str]
    photo_proof_residence: Any
    sex: Optional[str]

class Sep24TransactionRefundsPayment(BaseModel):
    id: str
    id_type: str
    amount: str
    fee: str

class Sep24TransactionRefunds(BaseModel):
    amount_refunded: str
    amount_fee: str
    payments: List[Sep24TransactionRefundsPayment]

class Sep24TransactionRequiredInfoUpdates(BaseModel):
    transaction: Any


Sep24TransactionStatus = Literal[
    "incomplete",
    "pending_user_transfer_start",
    "pending_user_transfer_complete",
    "pending_external",
    "pending_anchor",
    "pending_stellar",
    "pending_trust",
    "pending_user",
    "completed",
    "no_market",
    "too_small",
    "too_large",
    "error",
]

Sep24TransactionKind = Literal["deposit", "withdrawal"]

class Sep24Transaction(BaseModel):
    id: str
    kind: Sep24TransactionKind
    status: Sep24TransactionStatus
    status_eta: Optional[int]
    more_info_url: Optional[str]
    amount_in: Optional[str]
    amount_in_asset: Optional[str]
    amount_out: Optional[str]
    amount_out_asset: Optional[str]
    amount_fee: Optional[str]
    amount_fee_asset: Optional[str]
    from_address: Optional[str] = Field(None, alias="from")
    to_address: Optional[str] = Field(None, alias="to")
    external_extra: Optional[str]
    external_extra_text: Optional[str]
    deposit_memo: Optional[str]
    deposit_memo_type: Optional[str]
    withdraw_anchor_account: Optional[str]
    withdraw_memo: Optional[str]
    withdraw_memo_type: Optional[str]
    started_at: Optional[str]
    completed_at: Optional[str]
    stellar_transaction_id: Optional[str]
    external_transaction_id: Optional[str]
    message: Optional[str]
    refunds: Optional[Sep24TransactionRefunds]
    required_info_message: Optional[str]
    required_info_updates: Optional[Sep24TransactionRequiredInfoUpdates]
    claimable_balance_id: Optional[str]

    class Config:
        allow_population_by_field_name = True


class Asset(BaseModel):
    code: str
    issuer: Optional[str]
    decimal_places: int = 7

class Sep10GetRequest(BaseModel):
    """
    `GET <WEB_AUTH_ENDPOINT>` request schema
    """

    account: str
    home_domain: str
    memo: Optional[str] = None
    client_domain: Optional[str] = None


class Sep10GetResponse(BaseModel):
    """
    `GET <WEB_AUTH_ENDPOINT>` response schema
    """

    transaction: str
    network_passphrase: str


class Sep10PostRequest(BaseModel):
    """
    `POST <WEB_AUTH_ENDPOINT>` request schema
    """

    transaction: str


class Sep10PostResponse(BaseModel):
    """
    `POST <WEB_AUTH_ENDPOINT>` response schema
    """

    token: str


class Sep24DepositPostRequest(Sep9Customer, BaseModel):
    asset_code: str
    account: str
    asset_issuer: Optional[str]
    amount: Optional[str]
    memo_type: Optional[str]
    memo: Optional[str]
    wallet_name: Optional[str]
    wallet_url: Optional[str]
    lang: Optional[str] = "en"
    claimable_balance_supported: Optional[bool]

class Sep24WithdrawPostRequest(Sep9Customer, BaseModel):
    asset_code: str
    asset_issuer: Optional[str]
    amount: Optional[str]
    account: Optional[str]
    memo: Optional[str]
    memo_type: Optional[str]
    wallet_name: Optional[str]
    wallet_url: Optional[str]
    lang: Optional[str] = "en"

class Sep24DepositPostResponse(BaseModel):
    type: str = "interactive_customer_info_needed"
    url: str
    id: str

class Sep24PostResponse(BaseModel):
    type: str = "interactive_customer_info_needed"
    url: str
    id: str

class Sep24InfoRequest(BaseModel):
    lang: Optional[str] = "en"

class Sep24InfoResponseDeposit(BaseModel):
    enabled: bool
    min_amount: Optional[str]
    max_amount: Optional[str]
    fee_fixed: Optional[str]
    fee_percent: Optional[str]
    fee_minimum: Optional[str]

class Sep24InfoResponseWithdraw(BaseModel):
    enabled: bool
    min_amount: Optional[str]
    max_amount: Optional[str]
    fee_fixed: Optional[str]
    fee_percent: Optional[str]
    fee_minimum: Optional[str]

class Sep24InfoResponseFee(BaseModel):
    authentication_required: bool
    enabled: bool

class Sep24InfoResponseFeatures(BaseModel):
    account_creation: bool
    claimable_balances: bool

class Sep24InfoResponse(BaseModel):
    deposit: Dict[str, Sep24InfoResponseDeposit]
    withdraw: Dict[str, Sep24InfoResponseWithdraw]
    fee: Sep24InfoResponseFee
    features: Sep24InfoResponseFeatures

class Sep24FeeRequest(BaseModel):
    operation: str
    asset_code: str
    amount: str
    type: Optional[str]

class Sep24FeeResponse(BaseModel):
    operation: str
    asset_code: str
    amount: str
    type: Optional[str]

class Sep24TransactionsGetRequest(BaseModel):
    asset_code: str
    no_older_than: Optional[str]
    limit: Optional[int]
    kind: Optional[str]
    paging_id: Optional[str]

class Sep24TransactionsGetResponse(BaseModel):
    transactions: List[Sep24Transaction]

class Sep24TransactionGetRequest(BaseModel):
    id: str
    stellar_transaction_id: Optional[str]
    external_transaction_id: Optional[str]

class Sep24TransactionGetResponse(BaseModel):
    transaction: Sep24Transaction



from .exceptions import Sep10InvalidToken
from .models import (
    Asset,
    Sep9Customer,
    Sep10GetRequest,
    Sep10GetResponse,
    Sep10PostRequest,
    Sep10PostResponse,
    Sep24TransactionRefundsPayment,
    Sep24TransactionRefunds,
    Sep24TransactionRequiredInfoUpdates,
    Sep24TransactionKind,
    Sep24TransactionStatus,
    Sep24Transaction,
    Sep24DepositPostRequest,
    Sep24WithdrawPostRequest,
    Sep24DepositPostResponse,
    Sep24PostResponse,
    Sep24InfoRequest,
    Sep24InfoResponseDeposit,
    Sep24InfoResponseWithdraw,
    Sep24InfoResponseFee,
    Sep24InfoResponseFeatures,
    Sep24InfoResponse,
    Sep24FeeRequest,
    Sep24FeeResponse,
    Sep24TransactionsGetRequest,
    Sep24TransactionsGetResponse,
    Sep24TransactionGetRequest,
    Sep24TransactionGetResponse,
)
from .sep10 import Sep10, Sep10Token
from .sep24 import Sep24



import asyncio
from typing import Optional, Callable, Any, Union, List, Dict, Tuple
from abc import ABC, abstractmethod
import logging
from pydantic import BaseModel
from stellar_sdk.client.aiohttp_client import AiohttpClient
from stellar_sdk import ServerAsync, TransactionEnvelope
from stellar_sdk.transaction import Transaction as HorizonTransaction
from stellar_sdk.exceptions import (
    NotFoundError,
)
from stellar_sdk.xdr.utils import from_xdr_amount
from stellar_sdk.xdr import (
    PaymentResult,
    PathPaymentStrictSendResult,
    PathPaymentStrictReceiveResult,
    OperationResult,
    TransactionResult,
)
from stellar_sdk.operation import (
    Operation,
    Payment,
    PathPaymentStrictReceive,
    PathPaymentStrictSend,
)

from fawaris.models import (
    Sep9Customer,
    Sep24TransactionKind,
    Sep24TransactionStatus,
    Sep24Transaction,
    Sep24DepositPostRequest,
    Sep24WithdrawPostRequest,
    Sep24PostResponse,
    Sep24InfoRequest,
    Sep24InfoResponse,
    Sep24FeeRequest,
    Sep24FeeResponse,
    Sep24TransactionsGetRequest,
    Sep24TransactionsGetResponse,
    Sep24TransactionGetRequest,
    Sep24TransactionGetResponse,
    Asset,
)
from fawaris.sep10 import Sep10Token

PaymentOpResult = Union[
    PaymentResult, PathPaymentStrictSendResult, PathPaymentStrictReceiveResult
]
PaymentOp = Union[Payment, PathPaymentStrictReceive, PathPaymentStrictSend]


logger = logging.getLogger(__name__)


class Sep24(ABC):
    sep10_jwt_secret: str
    horizon_url: str
    network_passphrase: str
    assets: Dict[str, Asset]

    def __init__(
        self,
        sep10_jwt_secret: str,
        horizon_url: str,
        network_passphrase: str,
        assets: Dict[str, Asset],
    ):
        self.sep10_jwt_secret = sep10_jwt_secret
        self.horizon_url = horizon_url
        self.network_passphrase = network_passphrase
        self.assets = assets

    async def http_post_transactions_deposit_interactive(
        self, request: Sep24DepositPostRequest, token: Sep10Token
    ) -> Sep24PostResponse:
        info = await self.http_get_info(Sep24InfoRequest(lang=request.lang))
        try:
            if not info["deposit"][request.asset_code].enabled:
                raise KeyError()
        except KeyError:
            raise ValueError(f"Deposit is not enabled for asset {request.asset_code}")
        tx = await self.create_transaction(request, token)
        url = await self.get_interactive_url(request, token, tx)
        return Sep24PostResponse(
            url=url,
            id=tx.id,
        )

    async def http_post_transactions_withdraw_interactive(
        self, request: Sep24WithdrawPostRequest, token: Sep10Token
    ) -> Sep24PostResponse:
        info = await self.http_get_info(Sep24InfoRequest(lang=request.lang))
        try:
            if not info["withdraw"][request.asset_code].enabled:
                raise KeyError()
        except KeyError:
            raise ValueError(f"Withdrawal is not enabled for asset {request.asset_code}")
        tx = await self.create_transaction(request, token)
        url = await self.get_interactive_url(request, token, tx)
        return Sep24PostResponse(
            url=url,
            id=tx.id,
        )

    async def task_all(self) -> None:
        print("running task_all")
        coroutines = [
            self.task_poll_deposits_to_receive(),
            self.task_send_deposits(),
            self.task_poll_withdrawals_sent(),
            self.task_send_withdrawals(),
        ]
        results = await asyncio.gather(*coroutines)

    async def task_poll_deposits_to_receive(self) -> None:
        deposits_to_receive = await self.get_transactions(
            kind="deposit", status="pending_user_transfer_start"
        )
        print(f"deposits_to_receive: {len(deposits_to_receive)}")
        coroutines = [
            self.is_deposit_received(deposit) for deposit in deposits_to_receive
        ]
        results = await asyncio.gather(*coroutines, return_exceptions=True)
        received_deposits = []
        for deposit, result in zip(deposits_to_receive, results):
            if result is True:
                received_deposits.append(deposit)
            elif isinstance(result, Exception):
                logger.exception(result)
        await self.update_transactions(received_deposits, status="pending_anchor")

    async def task_send_deposits(self) -> None:
        deposits_received = await self.get_transactions(
            kind="deposit", status="pending_anchor"
        )
        print(f"deposits_received: {len(deposits_received)}")
        coroutines = [self.send_deposit(deposit) for deposit in deposits_received]
        results = await asyncio.gather(*coroutines, return_exceptions=True)
        for result in results:
            if isinstance(result, Exception):
                logger.exception(result)

    async def task_poll_withdrawals_sent(self) -> None:
        withdrawals_sent = await self.get_transactions(
            kind="withdrawal", status="pending_external"
        )
        print(f"withdrawals_sent: {len(withdrawals_sent)}")
        coroutines = [
            self.is_withdrawal_complete(withdrawal) for withdrawal in withdrawals_sent
        ]
        results = await asyncio.gather(*coroutines, return_exceptions=True)
        completed_withdrawals = []
        for withdrawal, result in zip(withdrawals_sent, results):
            if result is True:
                completed_withdrawals.append(withdrawal)
            elif isinstance(result, Exception):
                logger.exception(result)
        await self.update_transactions(completed_withdrawals, status="completed")

    async def task_send_withdrawals(self) -> None:
        withdrawals_received = await self.get_transactions(
            kind="withdrawal", status="pending_anchor"
        )
        print(f"withdrawals_received: {len(withdrawals_received)}")
        coroutines = [
            self.send_withdrawal(withdrawal) for withdrawal in withdrawals_received
        ]
        await asyncio.gather(*coroutines, return_exceptions=True)

    async def watch_withdrawals_to_receive(self) -> None:
        withdrawals_to_receive = await self.get_transactions(
            kind="withdrawal", status="pending_user_transfer_start"
        )
        accounts = set([withdrawal.withdraw_anchor_account for withdrawal in withdrawals_to_receive])
        await asyncio.gather(
            *[
                self.stream_withdraw_anchor_account(account)
                for account in accounts
            ],
        )

    async def stream_withdraw_anchor_account(self, account: str):
        async with ServerAsync(
            horizon_url=self.horizon_url, client=AiohttpClient()
        ) as server:
            try:
                # Ensure the distribution account actually exists
                await server.load_account(account)
            except NotFoundError:
                # This exception will crash the process, but the anchor needs
                # to provide valid accounts to watch.
                raise RuntimeError(
                    "Stellar distribution account does not exist in horizon"
                )
            cursor = self.get_withdraw_anchor_account_cursor(account)
            if cursor is None:
                cursor = "0"

            endpoint = server.transactions().for_account(account).cursor(cursor)
            async for response in endpoint.stream():
                try:
                    await self.process_stream_response(response, account)
                except Exception as e:
                    logger.exception(e)

    async def process_stream_response(self, response, account: str):
        # We should not match valid pending transactions with ones that were
        # unsuccessful on the stellar network. If they were unsuccessful, the
        # client is also aware of the failure and will likely attempt to
        # resubmit it, in which case we should match the resubmitted transaction
        if not response.get("successful"):
            return

        try:
            _ = response["id"]
            envelope_xdr = response["envelope_xdr"]
            memo = response["memo"]
            result_xdr = response["result_xdr"]
        except KeyError:
            return

        transactions = await self.get_transactions(
            kind="withdrawal",
            status="pending_user_transfer_start",
            memo=memo,
            withdraw_anchor_account=account,
        )

        if not transactions:
            return
        elif len(transactions) > 1:
            raise ValueError(f"Found multiple transactions matching memo: {memo}")
        transaction = transactions[0]

        #TODO check if tx hash has already been processed (relevant if cursor is 0)

        op_results = TransactionResult.from_xdr(result_xdr).result.results
        horizon_tx = TransactionEnvelope.from_xdr(
            envelope_xdr, network_passphrase=self.network_passphrase,
        ).transaction

        payment_data, source = await self.find_matching_payment_data(
            response, horizon_tx, op_results, transaction
        )
        if payment_data is None:
            logger.info(f"Transaction matching memo {memo} has no payment operation")
            return

        await self.process_withdrawal_received(
            transaction=transaction,
            amount_received=payment_data["amount"],
            from_address=source,
            horizon_response=response,
        )

    async def find_matching_payment_data(
        self,
        response: Dict,
        horizon_tx: HorizonTransaction,
        result_ops: List[OperationResult],
        transaction: Sep24Transaction,
    ) -> Optional[Dict]:
        matching_payment_data = None
        source = None
        ops = horizon_tx.operations
        for idx, op_result in enumerate(result_ops):
            op, op_result = await self.cast_operation_and_result(ops[idx], op_result)
            if not op_result:  # not a payment op
                continue
            maybe_payment_data = await self.check_for_payment_match(
                op, op_result, transaction.asset
            )
            if maybe_payment_data:
                if ops[idx].source:
                    _source = ops[idx].source.account_muxed or ops[idx].source.account_id
                else:
                    _source = (
                        horizon_tx.source.account_muxed or horizon_tx.source.account_id
                    )
                matching_payment_data = maybe_payment_data
                source = _source
                break

        return matching_payment_data, source

    async def cast_operation_and_result(
        self, operation: Operation, op_result: OperationResult
    ) -> Tuple[Optional[PaymentOp], Optional[PaymentOpResult]]:
        op_xdr_obj = operation.to_xdr_object()
        if isinstance(operation, Payment):
            return (
                Payment.from_xdr_object(op_xdr_obj),
                op_result.tr.payment_result,
            )
        elif isinstance(operation, PathPaymentStrictSend):
            return (
                PathPaymentStrictSend.from_xdr_object(op_xdr_obj),
                op_result.tr.path_payment_strict_send_result,
            )
        elif isinstance(operation, PathPaymentStrictReceive):
            return (
                PathPaymentStrictReceive.from_xdr_object(op_xdr_obj),
                op_result.tr.path_payment_strict_receive_result,
            )
        else:
            return None, None

    async def check_for_payment_match(
        self, operation: PaymentOp, op_result: PaymentOpResult, transaction: Sep24Transaction
    ) -> Optional[Dict]:
        payment_data = await self.get_payment_values(operation, op_result)
        #TODO add doc saying these fields need to be set when creating the tx
        asset = await self.get_transaction_asset(transaction)
        if (
            payment_data["destination"] == transaction.withdraw_anchor_account
            and payment_data["code"] == asset.code
            and payment_data["issuer"] == asset.issuer
        ):
            return payment_data
        else:
            return None

    async def get_payment_values(
        self, operation: PaymentOp, op_result: PaymentOpResult
    ) -> Dict:
        values = {
            "destination": operation.destination.account_id,
            "amount": None,
            "code": None,
            "issuer": None,
        }
        if isinstance(operation, Payment):
            values["amount"] = str(operation.amount)
            values["code"] = operation.asset.code
            values["issuer"] = operation.asset.issuer
        elif isinstance(operation, PathPaymentStrictSend):
            # since the dest amount is not specified in a strict-send op,
            # we need to get the dest amount from the operation's result
            #
            # this method of fetching amounts gives the "raw" amount, so
            # we need to divide by Operation._ONE: 10000000
            # (Stellar uses 7 decimals places of precision)
            values["amount"] = from_xdr_amount(op_result.success.last.amount.int64)
            values["code"] = operation.dest_asset.code
            values["issuer"] = operation.dest_asset.issuer
        elif isinstance(operation, PathPaymentStrictReceive):
            values["amount"] = str(operation.dest_amount)
            values["code"] = operation.dest_asset.code
            values["issuer"] = operation.dest_asset.issuer
        else:
            raise ValueError("Unexpected operation, expected payment or path payment")
        return values

    @abstractmethod
    async def http_get_info(self, request: Sep24InfoRequest) -> Sep24InfoResponse:
        raise NotImplementedError()

    @abstractmethod
    async def http_get_fee(
        self, request: Sep24FeeRequest, token: Optional[Sep10Token] = None
    ) -> Sep24FeeResponse:
        raise NotImplementedError()

    @abstractmethod
    async def http_get_transactions(
        self, request: Sep24TransactionsGetRequest, token: Sep10Token
    ) -> Sep24TransactionsGetResponse:
        raise NotImplementedError()

    @abstractmethod
    async def http_get_transaction(
        self, request: Sep24TransactionGetRequest, token: Sep10Token
    ) -> Sep24TransactionGetResponse:
        raise NotImplementedError()

    @abstractmethod
    async def create_transaction(
        self,
        request: Union[Sep24DepositPostRequest, Sep24WithdrawPostRequest],
        token: Sep10Token,
    ) -> Sep24Transaction:
        raise NotImplementedError()

    @abstractmethod
    async def get_interactive_url(
        self,
        request: Union[Sep24DepositPostRequest, Sep24WithdrawPostRequest],
        token: Sep10Token,
        tx: Sep24Transaction,
    ) -> str:
        raise NotImplementedError()

    @abstractmethod
    async def get_transactions(self,
        kind: Optional[Sep24TransactionKind] = None,
        status: Optional[Sep24TransactionStatus] = None,
        memo: Optional[str] = None,
        withdraw_anchor_account: Optional[str] = None,
    ) -> List[Sep24Transaction]:
        raise NotImplementedError()

    @abstractmethod
    async def get_transaction_asset(self, transaction: Sep24Transaction) -> Asset:
        raise NotImplementedError()

    @abstractmethod
    async def is_deposit_received(self, deposit: Sep24Transaction) -> bool:
        raise NotImplementedError()

    @abstractmethod
    async def is_withdrawal_complete(self, withdrawal: Sep24Transaction) -> bool:
        raise NotImplementedError()

    @abstractmethod
    async def process_withdrawal_received(self,
        transaction: Sep24Transaction,
        amount_received: str,
        from_address: str,
        horizon_response: Dict,
    ):
        raise NotImplementedError()

    @abstractmethod
    async def update_transactions(self, transactions: List[Sep24Transaction], **values):
        raise NotImplementedError()

    @abstractmethod
    async def send_withdrawal(self, withdrawal: Sep24Transaction) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def send_deposit(self, deposit: Sep24Transaction) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def get_withdraw_anchor_account_cursor(self, account: str) -> Optional[str]:
        raise NotImplementedError()




from typing import Optional, List, Dict, Union, Iterable, Callable, Any
from urllib.parse import urlparse
from datetime import datetime, timezone
import logging
import os
import os.path

import jwt
from jwt import decode
from jwt.exceptions import InvalidTokenError
import toml
from stellar_sdk.sep.stellar_web_authentication import (
    build_challenge_transaction,
    read_challenge_transaction,
    verify_challenge_transaction_threshold,
    verify_challenge_transaction_signed_by_client_master_key,
)
from stellar_sdk.operation import ManageData
from stellar_sdk import ServerAsync, Keypair, MuxedAccount
from stellar_sdk.sep.stellar_toml import fetch_stellar_toml
from stellar_sdk.sep.exceptions import (
    InvalidSep10ChallengeError,
    StellarTomlNotFoundError,
)
from stellar_sdk.exceptions import (
    NotFoundError,
    ConnectionError,
    Ed25519PublicKeyInvalidError,
    Ed25519SecretSeedInvalidError,
    MuxedEd25519AccountInvalidError,
    ValueError as StellarSdkValueError,
)
from stellar_sdk.client.aiohttp_client import AiohttpClient
from stellar_sdk.strkey import StrKey

from fawaris.models import (
    Sep10GetRequest,
    Sep10GetResponse,
    Sep10PostRequest,
    Sep10PostResponse,
)
from fawaris.exceptions import Sep10InvalidToken

logger = logging.getLogger(__name__)

class Sep10:
    host_url: str
    home_domains: Union[str, Iterable[str]]
    horizon_url: str
    network_passphrase: str
    signing_secret: str
    jwt_secret: str
    client_domain_required: bool
    client_domains_allowed: Optional[List[str]]
    client_domains_denied: Optional[List[str]]

    server_account_id: str
    web_auth_domain: str

    def __init__(
        self,
        host_url: str,
        home_domains: Union[str, Iterable[str]],
        horizon_url: str,
        network_passphrase: str,
        signing_secret: str,
        jwt_secret: str,
        client_domain_required: bool = False,
        client_domains_allowed: Optional[List[str]] = None,
        client_domains_denied: Optional[List[str]] = None,
    ):
        """
        Implementation of `SEP0010 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md>`_

        :param host_url: URL to this Anchor instance, ex: https://myanchor.domain.com
        :param home_domains: The home domain that is expected to be included
            in the first Manage Data operation's string key. If a list is
            provided, one of the domain names in the array must match.
        :param horizon_url: Stellar API url, ex: https://horizon-testnet.stellar.org
        :param network_passphrase: Network passphrase, must be the same as
            NETWORK_PASSPHRASE from this Anchor's stellar.toml
        :param signing_secret: Challenge transaction signing seed, which is the
            secret key of the public key SIGNING_KEY from this Anchor's
            stellar.toml
        :param jwt_secret: JWT secret key used to encode the JWT
        :param client_domain_required: Require client_domain when building
            challenge transaction
        :param client_domains_allowed: List of allowed client_domain values.
            If not set, any client_domain is accepted
        :param client_domains_denied: List of denied client_domain values.
            If not set, any client_domain is accepted. If a client_domain is
            listed both here and in client_domains_allowed, it will be denied
        """
        if not urlparse(host_url).netloc:
            raise ValueError(f"{host_url} is not a valid host_url")
        self.host_url = host_url
        self.web_auth_domain = urlparse(host_url).netloc

        self.home_domains = home_domains
        self.horizon_url = horizon_url
        self.network_passphrase = network_passphrase

        try:
            kp = Keypair.from_secret(signing_secret)
        except Ed25519SecretSeedInvalidError:
            raise ValueError("'signing_secret' contains an invalid secret key")
        self.signing_secret = signing_secret
        self.server_account_id = kp.public_key

        self.jwt_secret = jwt_secret
        self.client_domain_required = client_domain_required
        self.client_domains_allowed = client_domains_allowed
        self.client_domains_denied = client_domains_denied

    async def http_get(
        self,
        request: Sep10GetRequest,
        timeout: int = 900,
    ) -> Sep10GetResponse:
        memo = request.memo
        if memo:
            try:
                memo = int(memo)
            except ValueError:
                raise ValueError("invalid 'memo' value. Expected a 64-bit integer.")
            if request.account.startswith("M"):
                raise ValueError(
                    "'memo' cannot be passed with a muxed client account "
                    "address (M...)"
                )
        else:
            memo = None

        if request.home_domain not in self.home_domains:
            raise ValueError(
                "invalid 'home_domain' value. Accepted values: " f"{self.home_domains}"
            )

        if urlparse(f"https://{request.home_domain}").netloc != request.home_domain:
            raise ValueError("'home_domain' must be a valid hostname")

        if request.client_domain:
            if (
                urlparse(f"https://{request.client_domain}").netloc
                != request.client_domain
            ):
                raise ValueError("'client_domain' must be a valid hostname")

            if (
                self.client_domains_denied is not None
                and request.client_domain in self.client_domains_denied
            ):
                raise ValueError("'client_domain' value is denied")

            if (
                self.client_domains_allowed is not None
                and request.client_domain not in self.client_domains_allowed
            ):
                raise ValueError("'client_domain' value is not allowed")

            try:
                client_signing_key = await self._get_client_signing_key(
                    request.client_domain
                )
            except (
                ConnectionError,
                StellarTomlNotFoundError,
                toml.decoder.TomlDecodeError,
            ):
                raise ValueError("unable to fetch 'client_domain' SIGNING_KEY")

        else:
            if self.client_domain_required:
                raise ValueError("'client_domain' is required")
            client_domain = None
            client_signing_key = None

        transaction = build_challenge_transaction(
            server_secret=self.signing_secret,
            client_account_id=request.account,
            home_domain=request.home_domain,
            web_auth_domain=self.web_auth_domain,
            network_passphrase=self.network_passphrase,
            timeout=timeout,
            client_domain=request.client_domain,
            client_signing_key=client_signing_key,
            memo=memo,
        )
        return Sep10GetResponse(
            transaction=transaction,
            network_passphrase=self.network_passphrase,
        )

    async def http_post(self, request: Sep10PostRequest) -> Sep10PostResponse:
        client_domain = await self._validate_challenge_xdr(request)
        return Sep10PostResponse(token=self._generate_jwt(request, client_domain))

    def _get_http_client(self, request_timeout=11):
        return AiohttpClient(request_timeout=request_timeout)

    async def _get_client_signing_key(self, client_domain):
        client_toml_contents = await fetch_stellar_toml(
            client_domain,
            client=self._get_http_client(),
        )
        client_signing_key = client_toml_contents.get("SIGNING_KEY")
        if not client_signing_key:
            raise ValueError("SIGNING_KEY not present on 'client_domain' TOML")
        try:
            Keypair.from_public_key(client_signing_key)
        except Ed25519PublicKeyInvalidError:
            raise ValueError("invalid SIGNING_KEY value on 'client_domain' TOML")
        return client_signing_key

    async def _validate_challenge_xdr(self, request: Sep10GetRequest):
        logger.debug("Validating challenge transaction")
        try:
            challenge = read_challenge_transaction(
                challenge_transaction=request.transaction,
                server_account_id=self.server_account_id,
                home_domains=self.home_domains,
                web_auth_domain=self.web_auth_domain,
                network_passphrase=self.network_passphrase,
            )
        except (InvalidSep10ChallengeError, TypeError) as e:
            raise ValueError(e)

        client_domain = None
        for operation in challenge.transaction.transaction.operations:
            if (
                isinstance(operation, ManageData)
                and operation.data_name == "client_domain"
            ):
                client_domain = operation.data_value.decode()
                break

        # extract the Stellar account from the muxed account to check for its existence
        stellar_account = challenge.client_account_id
        if challenge.client_account_id.startswith("M"):
            stellar_account = MuxedAccount.from_account(
                challenge.client_account_id
            ).account_id

        try:
            async with ServerAsync(
                horizon_url=self.horizon_url, client=self._get_http_client()
            ) as server:
                account = await server.load_account(stellar_account)
        except NotFoundError:
            logger.debug("Account does not exist, using client's master key to verify")
            try:
                verify_challenge_transaction_signed_by_client_master_key(
                    challenge_transaction=request.transaction,
                    server_account_id=self.server_account_id,
                    home_domains=self.home_domains,
                    web_auth_domain=self.web_auth_domain,
                    network_passphrase=self.network_passphrase,
                )
                if (client_domain and len(challenge.transaction.signatures) != 3) or (
                    not client_domain and len(challenge.transaction.signatures) != 2
                ):
                    raise InvalidSep10ChallengeError(
                        "There is more than one client signer on a challenge "
                        "transaction for an account that doesn't exist"
                    )
            except InvalidSep10ChallengeError as e:
                raise ValueError(
                    f"Missing or invalid signature(s) for {challenge.client_account_id}: {str(e)}"
                )
            else:
                logger.debug("Challenge verified using client's master key")
                return client_domain

        signers = account.load_ed25519_public_key_signers()
        threshold = account.thresholds.med_threshold
        signers_found = verify_challenge_transaction_threshold(
            challenge_transaction=request.transaction,
            server_account_id=self.server_account_id,
            home_domains=self.home_domains,
            web_auth_domain=self.web_auth_domain,
            network_passphrase=self.network_passphrase,
            threshold=threshold,
            signers=signers,
        )
        logger.debug(
            f"Challenge verified using account signers: {[s.account_id for s in signers_found]}"
        )

        return client_domain

    def _generate_jwt(
        self, request: Sep10PostRequest, client_domain: str = None
    ) -> str:
        challenge = read_challenge_transaction(
            challenge_transaction=request.transaction,
            server_account_id=self.server_account_id,
            home_domains=self.home_domains,
            web_auth_domain=self.web_auth_domain,
            network_passphrase=self.network_passphrase,
        )
        logger.debug(f"Generating SEP-10 token for account {challenge.client_account_id}")

        # set iat value to minimum timebound of the challenge so that the JWT returned
        # for a given challenge is always the same.
        # https://github.com/stellar/stellar-protocol/pull/982
        issued_at = challenge.transaction.transaction.time_bounds.min_time

        # format sub value based on muxed account or memo
        if challenge.client_account_id.startswith("M") or not challenge.memo:
            sub = challenge.client_account_id
        else:
            sub = f"{challenge.client_account_id}:{challenge.memo}"

        jwt_dict = {
            "iss": os.path.join(self.host_url, "auth"),
            "sub": sub,
            "iat": issued_at,
            "exp": issued_at + 24 * 60 * 60,
            "jti": challenge.transaction.hash().hex(),
            "client_domain": client_domain,
        }
        return jwt.encode(jwt_dict, self.jwt_secret, algorithm="HS256")

class Sep10Token:
    """
    SEP-10 token representation.
    See https://github.com/stellar/django-polaris/blob/v2.2.0/polaris/polaris/sep10/token.py
    """

    _REQUIRED_FIELDS = {"iss", "sub", "iat", "exp"}

    def __init__(self, jwt: Union[str, Dict], jwt_secret):
        if isinstance(jwt, str):
            try:
                jwt = decode(jwt, jwt_secret, algorithms=["HS256"])
            except InvalidTokenError as e:
                raise Sep10InvalidToken("unable to decode jwt" + str(e))
        elif not isinstance(jwt, Dict):
            raise Sep10InvalidToken(
                "invalid type for 'jwt' parameter: must be a string or dictionary"
            )

        if not self._REQUIRED_FIELDS.issubset(set(jwt.keys())):
            raise Sep10InvalidToken(
                f"jwt is missing one of the required fields: {', '.join(self._REQUIRED_FIELDS)}"
            )

        memo = None
        stellar_account = None
        if jwt["sub"].startswith("M"):
            try:
                StrKey.decode_muxed_account(jwt["sub"])
            except (MuxedEd25519AccountInvalidError, StellarSdkValueError):
                raise Sep10InvalidToken(f"invalid muxed account address: {jwt['sub']}")
        elif ":" in jwt["sub"]:
            try:
                stellar_account, memo = jwt["sub"].split(":")
            except ValueError:
                raise Sep10InvalidToken(f"improperly formatted 'sub' value: {jwt['sub']}")
        else:
            stellar_account = jwt["sub"]

        if stellar_account:
            try:
                Keypair.from_public_key(stellar_account)
            except Ed25519PublicKeyInvalidError:
                raise Sep10InvalidToken(f"invalid Stellar public key: {jwt['sub']}")

        if memo:
            try:
                int(memo)
            except ValueError:
                raise Sep10InvalidToken(
                    f"invalid memo in 'sub' value, expected 64-bit integer: {memo}"
                )

        try:
            iat = datetime.fromtimestamp(jwt["iat"], tz=timezone.utc)
        except (OSError, ValueError, OverflowError):
            raise Sep10InvalidToken("invalid iat value")
        try:
            exp = datetime.fromtimestamp(jwt["exp"], tz=timezone.utc)
        except (OSError, ValueError, OverflowError):
            raise Sep10InvalidToken("invalid exp value")

        now = datetime.now(tz=timezone.utc)
        if now < iat or now > exp:
            raise Sep10InvalidToken("jwt is no longer valid")

        client_domain = jwt.get("client_domain")
        if (
            client_domain
            and urlparse(f"https://{client_domain}").netloc != client_domain
        ):
            raise Sep10InvalidToken("'client_domain' must be a hostname")

        self._payload = jwt

    @property
    def account(self) -> str:
        """
        The Stellar account (`G...`) authenticated. Note that a muxed account
        could have been authenticated, in which case `Token.muxed_account` should
        be used.
        """
        if self._payload["sub"].startswith("M"):
            return MuxedAccount.from_account(self._payload["sub"]).account_id
        elif ":" in self._payload["sub"]:
            return self._payload["sub"].split(":")[0]
        else:
            return self._payload["sub"]

    @property
    def muxed_account(self) -> Optional[str]:
        """
        The M-address specified in the payload's ``sub`` value, if present
        """
        return self._payload["sub"] if self._payload["sub"].startswith("M") else None

    @property
    def memo(self) -> Optional[int]:
        """
        The memo included with the payload's ``sub`` value, if present
        """
        return (
            int(self._payload["sub"].split(":")[1])
            if ":" in self._payload["sub"]
            else None
        )

    @property
    def issuer(self) -> str:
        """
        The principal that issued a token, RFC7519, Section 4.1.1 — a Uniform
        Resource Identifier (URI) for the issuer
        (https://example.com or https://example.com/G...)
        """
        return self._payload["iss"]

    @property
    def issued_at(self) -> datetime:
        """
        The time at which the JWT was issued RFC7519, Section 4.1.6 -
        represented as a UTC datetime object
        """
        return datetime.fromtimestamp(self._payload["iat"], tz=timezone.utc)

    @property
    def expires_at(self) -> datetime:
        """
        The expiration time on or after which the JWT will not accepted for
        processing, RFC7519, Section 4.1.4 — represented as a UTC datetime object
        """
        return datetime.fromtimestamp(self._payload["exp"], tz=timezone.utc)

    @property
    def client_domain(self) -> Optional[str]:
        """
        A nonstandard JWT claim containing the client's home domain, included if
        the challenge transaction contained a ``client_domain`` ManageData operation
        """
        return self._payload.get("client_domain")

    @property
    def payload(self) -> dict:
        """
        The decoded contents of the JWT string
        """
        return self._payload



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fawaris']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.1,<3.0',
 'pydantic>=1.0,<2.0',
 'stellar-sdk>=7.0.1,<8.0.0',
 'typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'fawaris',
    'version': '0.1.37',
    'description': 'Async framework-agnostic Stellar Anchor interface',
    'long_description': None,
    'author': 'Yuri Escalianti',
    'author_email': 'yuriescl@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yuriescl/fawaris-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
